@@ -1,8 +1,10 @@
-import click
-from pathlib import Path
-import typing
-from ..utils.printer import print_result
 import os
+import typing
+from pathlib import Path
+
+import click
+
+from ..utils.printer import print_result
 
 if typing.TYPE_CHECKING:
     from ..client import Primitive
@@ -39,11 +41,10 @@ def download_ref_command(
     destination: Path = Path.cwd(),
 ):
     primitive: Primitive = context.obj.get("PRIMITIVE")
-    result, message = primitive.git.download_git_repository_at_ref(
+    path = primitive.git.download_git_repository_at_ref(
         git_repo_full_name=git_repo_full_name,
         git_ref=git_ref,
         github_access_token=github_access_token,
         destination=destination,
     )
-    fg = "red" if not result else "green"
-    print_result(message, context=context, fg=fg)
+    print_result(message=path, context=context)
