@@ -2,35 +2,49 @@ import csv
 import io
 import json
 import platform
-from shutil import which
 import subprocess
-from typing import Dict, List, Optional
-import click
-from loguru import logger
-from primitive.utils.memory_size import MemorySize
-from gql import gql
-from aiohttp import client_exceptions
-from ..utils.config import update_config_file
-from ..utils.auth import guard
-
 import typing
+from shutil import which
+from typing import Dict, List, Optional
+
+import click
+from aiohttp import client_exceptions
+from gql import gql
+from loguru import logger
+
+from primitive.graphql.relay import from_base64
+from primitive.utils.memory_size import MemorySize
+
+from ..utils.auth import guard
+from ..utils.config import update_config_file
+from ..utils.exceptions import P_CLI_100
+from .graphql.mutations import (
+    hardware_checkin_mutation,
+    hardware_update_mutation,
+    register_child_hardware_mutation,
+    register_hardware_mutation,
+    unregister_hardware_mutation,
+)
+from .graphql.queries import (
+    hardware_details,
+    hardware_list,
+    nested_children_hardware_list,
+)
 
 if typing.TYPE_CHECKING:
     pass
 
 
+from primitive.hardware.android import AndroidDevice, list_devices
 from primitive.utils.actions import BaseAction
+from primitive.utils.shell import does_executable_exist
 
 
 class Hardware(BaseAction):
     def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
-        self.previous_state = {
-            "isHealthy": False,
-            "isQuarantined": False,
-            "isAvailable": False,
-            "isOnline": False,
-        }
+        self.status_cache = {}
+        self.children = []
 
     def _get_darwin_system_profiler_values(self) -> Dict[str, str]:
         system_profiler_hardware_data_type = subprocess.check_output(
@@ -63,7 +77,7 @@ class Hardware(BaseAction):
                 )
             except subprocess.CalledProcessError as exception:
                 message = f"Error running system_profiler: {exception}"
-                logger.error(message)
+                logger.exception(message)
                 return supported_metal_device
 
             try:
@@ -72,7 +86,7 @@ class Hardware(BaseAction):
                 )
             except json.JSONDecodeError as exception:
                 message = f"Error decoding JSON: {exception}"
-                logger.error(message)
+                logger.exception(message)
                 return supported_metal_device
 
             # Checks if any attached displays have metal support
@@ -103,7 +117,7 @@ class Hardware(BaseAction):
                 )
             except subprocess.CalledProcessError as exception:
                 message = f"Command {nvidia_smi_query_gpu_csv_command} failed with exception: {exception}"  # noqa
-                logger.error(message)
+                logger.exception(message)
                 raise exception
 
             try:
@@ -115,7 +129,7 @@ class Hardware(BaseAction):
                 )
             except UnicodeDecodeError as exception:
                 message = f"Error decoding: {exception}"
-                logger.error(message)
+                logger.exception(message)
                 raise exception
 
             nvidia_smi_query_gpu_csv_dict_reader = csv.DictReader(
@@ -146,7 +160,7 @@ class Hardware(BaseAction):
                     )
                 except subprocess.CalledProcessError as exception:
                     message = f"Error running {system_profiler_hardware_data_type_command}: {exception}"  # noqa
-                    logger.error(message)
+                    logger.exception(message)
                     raise exception
 
                 try:
@@ -155,7 +169,7 @@ class Hardware(BaseAction):
                     )
                 except json.JSONDecodeError as exception:
                     message = f"Error decoding JSON: {exception}"  # noqa
-                    logger.error(message)
+                    logger.exception(message)
                     raise exception
 
                 metal_device_json = system_profiler_hardware_data_type_json[
@@ -267,40 +281,25 @@ class Hardware(BaseAction):
         return system_info
 
     @guard
-    def register(self):
+    def register(self, organization_id: Optional[str] = None):
         system_info = self.get_system_info()
-        mutation = gql(
-            """
-            mutation registerHardware($input: RegisterHardwareInput!) {
-                registerHardware(input: $input) {
-                    ... on Hardware {
-                        fingerprint
-                    }
-                    ... on OperationInfo {
-                          messages {
-                            kind
-                            message
-                            field
-                            code
-                        }
-                    }
-                }
-            }
-        """
-        )
+        mutation = gql(register_hardware_mutation)
         input = {"systemInfo": system_info}
+        if organization_id:
+            input["organizationId"] = organization_id
         variables = {"input": input}
-        result = self.primitive.session.execute(mutation, variable_values=variables)
-        if messages := result.get("registerHardware").get("messages"):
+        result = self.primitive.session.execute(
+            mutation, variable_values=variables, get_execution_result=True
+        )
+        if messages := result.data.get("registerHardware").get("messages"):
             for message in messages:
-                logger.enable("primitive")
                 if message.get("kind") == "ERROR":
                     logger.error(message.get("message"))
                 else:
                     logger.debug(message.get("message"))
             return False
 
-        fingerprint = result.get("registerHardware").get("fingerprint")
+        fingerprint = result.data.get("registerHardware").get("fingerprint")
 
         self.primitive.host_config["fingerprint"] = fingerprint
         self.primitive.full_config[self.primitive.host] = self.primitive.host_config
@@ -310,7 +309,30 @@ class Hardware(BaseAction):
         # and headers are set correctly
         self.primitive.get_host_config()
         self.check_in_http(is_healthy=True)
-        return True
+        for child in self._list_local_children():
+            self.register_child(child=child)
+        return result
+
+    @guard
+    def unregister(self, organization_id: Optional[str] = None):
+        mutation = gql(unregister_hardware_mutation)
+        input = {
+            "fingerprint": self.primitive.host_config.get("fingerprint"),
+        }
+        variables = {"input": input}
+        result = self.primitive.session.execute(
+            mutation, variable_values=variables, get_execution_result=True
+        )
+
+        if messages := result.data.get("unregisterHardware").get("messages"):
+            for message in messages:
+                if message.get("kind") == "ERROR":
+                    logger.error(message.get("message"))
+                else:
+                    logger.debug(message.get("message"))
+            return False
+
+        return result
 
     @guard
     def update_hardware_system_info(self):
@@ -335,36 +357,20 @@ class Hardware(BaseAction):
             "systemInfo": system_info,
         }
 
-        mutation = gql(
-            """
-            mutation hardwareUpdate($input: HardwareUpdateInput!) {
-                hardwareUpdate(input: $input) {
-                    ... on Hardware {
-                        systemInfo
-                    }
-                    ... on OperationInfo {
-                          messages {
-                            kind
-                            message
-                            field
-                            code
-                        }
-                    }
-                }
-            }
-            """
-        )
+        mutation = gql(hardware_update_mutation)
 
         input = new_state
         variables = {"input": input}
         try:
-            result = self.primitive.session.execute(mutation, variable_values=variables)
+            result = self.primitive.session.execute(
+                mutation, variable_values=variables, get_execution_result=True
+            )
         except client_exceptions.ClientConnectorError as exception:
-            message = " [*] Failed to update hardware system info! "
-            logger.error(message)
+            message = "Failed to update hardware system info! "
+            logger.exception(message)
             raise exception
 
-        message = " [*] Updated hardware system info successfully! "
+        message = "Updated hardware system info successfully! "
         logger.info(message)
 
         return result
@@ -376,8 +382,13 @@ class Hardware(BaseAction):
         is_quarantined: bool = False,
         is_available: bool = False,
         is_online: bool = True,
+        fingerprint: Optional[str] = None,
+        stopping_agent: Optional[bool] = False,
     ):
-        fingerprint = self.primitive.host_config.get("fingerprint", None)
+        # if no fingerprint supplied from argument try from the host_config
+        if not fingerprint:
+            fingerprint = self.primitive.host_config.get("fingerprint", None)
+
         if not fingerprint:
             message = (
                 "No fingerprint found. Please register: primitive hardware register"
@@ -391,28 +402,10 @@ class Hardware(BaseAction):
             "isOnline": is_online,
         }
 
-        mutation = gql(
-            """
-            mutation checkIn($input: CheckInInput!) {
-                checkIn(input: $input) {
-                    ... on Hardware {
-                        createdAt
-                        updatedAt
-                        lastCheckIn
-                    }
-                    ... on OperationInfo {
-                          messages {
-                            kind
-                            message
-                            field
-                            code
-                        }
-                    }
-                }
-            }
-        """  # noqa
-        )
-        input = new_state
+        mutation = gql(hardware_checkin_mutation)
+        input = {**new_state}
+        if fingerprint:
+            input["fingerprint"] = fingerprint
         variables = {"input": input}
         try:
             result = self.primitive.session.execute(
@@ -421,19 +414,29 @@ class Hardware(BaseAction):
             checkin_success = result.data.get("checkIn").get("lastCheckIn")
             if messages := result.data.get("checkIn").get("messages"):
                 for message in messages:
-                    logger.enable("primitive")
                     if message.get("kind") == "ERROR":
                         logger.error(message.get("message"))
                     else:
                         logger.debug(message.get("message"))
 
             if checkin_success:
-                previous_state = self.previous_state
-                self.previous_state = new_state.copy()
+                if fingerprint in self.status_cache:
+                    previous_status = self.status_cache[fingerprint]
+                else:
+                    logger.error("No previous status found.")
+                    previous_status = {
+                        "isHealthy": False,
+                        "isQuarantined": False,
+                        "isAvailable": False,
+                        "isOnline": False,
+                    }
+                self.status_cache[fingerprint] = new_state.copy()
 
-                message = " [*] Checked in successfully: "
+                message = f"Checked in successfully for {fingerprint}: "
+                is_new_status = False
                 for key, value in new_state.items():
-                    if value != previous_state.get(key, None):
+                    if value != previous_status.get(key, None):
+                        is_new_status = True
                         if value is True:
                             message = (
                                 message
@@ -448,81 +451,219 @@ class Hardware(BaseAction):
                                 + click.style("✅")
                                 + click.style(" ==> 💤 ", fg="yellow")
                             )
+                if is_new_status is False:
+                    message += "No changes."
                 logger.info(message)
             else:
                 message = "Failed to check in!"
                 raise Exception(message)
             return result
         except client_exceptions.ClientConnectorError as exception:
-            message = " [*] Failed to check in! "
-            logger.error(message)
-            raise exception
+            if not stopping_agent:
+                message = "Failed to check in! "
+                logger.error(message)
+                raise exception
+            else:
+                raise P_CLI_100
 
     @guard
-    def get_hardware_list(self, fingerprint: Optional[str] = None):
-        query = gql(
-            """
-fragment HardwareFragment on Hardware {
-  id
-  pk
-  name
-  slug
-  createdAt
-  updatedAt
-  isAvailable
-  isOnline
-  isQuarantined
-  isHealthy
-  capabilities {
-    id
-    pk
-  }
-  activeReservation {
-    id
-    pk
-  }
-}
+    def get_hardware_list(
+        self,
+        fingerprint: Optional[str] = None,
+        id: Optional[str] = None,
+        slug: Optional[str] = None,
+        nested_children: Optional[bool] = False,
+    ):
+        query = gql(nested_children_hardware_list if nested_children else hardware_list)
 
-query hardwareList(
-  $before: String
-  $after: String
-  $first: Int
-  $last: Int
-  $filters: HardwareFilters
-) {
-  hardwareList(
-    before: $before
-    after: $after
-    first: $first
-    last: $last
-    filters: $filters
-  ) {
-    totalCount
-    edges {
-      cursor
-      node {
-        ...HardwareFragment
-      }
-    }
-  }
-}
-"""
+        filters = {
+            "isRegistered": {"exact": True},
+        }
+        if fingerprint is not None:
+            filters["fingerprint"] = {"exact": fingerprint}
+        if slug is not None:
+            filters["slug"] = {"exact": slug}
+        if id is not None:
+            filters["id"] = {"exact": id}
+        if nested_children is True:
+            filters["hasParent"] = {"exact": False}
+
+        variables = {
+            "filters": filters,
+        }
+        result = self.primitive.session.execute(
+            query, variable_values=variables, get_execution_result=True
         )
+        return result
+
+    @guard
+    def get_hardware_details(
+        self,
+        fingerprint: Optional[str] = None,
+        id: Optional[str] = None,
+        slug: Optional[str] = None,
+    ):
+        query = gql(hardware_details)
 
         filters = {}
-        if fingerprint:
+        if fingerprint is not None:
             filters["fingerprint"] = {"exact": fingerprint}
+        if slug is not None:
+            filters["slug"] = {"exact": slug}
+        if id is not None:
+            filters["id"] = {"exact": id}
 
         variables = {
             "first": 1,
             "filters": filters,
         }
-        result = self.primitive.session.execute(query, variable_values=variables)
-        return result
+        result = self.primitive.session.execute(
+            query, variable_values=variables, get_execution_result=True
+        )
+        if edges := result.data["hardwareList"]["edges"]:
+            return edges[0]["node"]
+        else:
+            raise Exception(f"No hardware found with {filters}")
 
     def get_own_hardware_details(self):
-        hardware_list_data = self.get_hardware_list(
-            fingerprint=self.primitive.host_config.get("fingerprint")
+        hardware_list_result = self.get_hardware_list(
+            fingerprint=self.primitive.host_config.get("fingerprint"),
+            nested_children=True,
+        )
+        hardware = (
+            hardware_list_result.data.get("hardwareList").get("edges")[0].get("node")
         )
-        hardware = hardware_list_data.get("hardwareList").get("edges")[0].get("node")
         return hardware
+
+    def get_hardware_from_slug_or_id(self, hardware_identifier: str):
+        is_id = False
+        is_slug = False
+        id = None
+        # first check if the hardware_identifier is a slug or ID
+        try:
+            type_name, id = from_base64(hardware_identifier)
+
+            is_id = True
+            if type_name == "Hardware":
+                is_hardware_type = True
+            else:
+                raise Exception(
+                    f"ID was not for Hardware, you supplied an ID for a {type_name}"
+                )
+
+        except ValueError:
+            is_slug = True
+
+        hardware = None
+
+        if is_id and is_hardware_type:
+            hardware = self.get_hardware_details(id=id)
+        elif is_slug:
+            hardware = self.get_hardware_details(slug=hardware_identifier)
+
+        return hardware
+
+    @guard
+    def register_child(self, child: AndroidDevice):
+        system_info = child.system_info
+        mutation = gql(register_child_hardware_mutation)
+        input = {"childSystemInfo": system_info}
+        variables = {"input": input}
+        result = self.primitive.session.execute(
+            mutation, variable_values=variables, get_execution_result=True
+        )
+
+        if messages := result.data.get("registerChildHardware").get("messages"):
+            for message in messages:
+                if message.get("kind") == "ERROR":
+                    logger.error(message.get("message"))
+                else:
+                    logger.debug(message.get("message"))
+            return False
+        return result
+
+    def _list_local_children(self) -> List[AndroidDevice]:
+        if does_executable_exist("adb"):
+            self.children: List[AndroidDevice] = list_devices()
+        return self.children
+
+    @guard
+    def _remove_child(self):
+        pass
+
+    @guard
+    def _sync_children(self, hardware: Optional[Dict[str, str]] = None):
+        # get the existing children if any from the hardware details
+        # get the latest children from the node
+        # compare the two and update the node with the latest children
+        # remove any children from remote that are not in the latest children
+        if not hardware:
+            hardware = self.primitive.hardware.get_own_hardware_details()
+        if not hardware:
+            logger.error("No hardware found.")
+            return
+
+        remote_children = hardware.get("children", [])
+        local_children = self.primitive.hardware._list_local_children()
+
+        new_child_registered = False
+        # need to register new children that were not previously registered
+        if len(remote_children) < len(local_children):
+            logger.info("New children found.")
+            for local_child in local_children:
+                found = False
+                for remote_child in remote_children:
+                    if remote_child["slug"] == local_child.slug:
+                        found = True
+                        break
+                if not found:
+                    try:
+                        logger.info(f"Registering new child: {local_child.slug}")
+                        self.primitive.hardware.register_child(child=local_child)
+                        new_child_registered = True
+                    except Exception as exception:
+                        logger.exception(f"Error registering new children: {exception}")
+
+        # if a new child was registered, get the fresh state of the world
+        if new_child_registered:
+            hardware = self.primitive.hardware.get_own_hardware_details()
+            remote_children = hardware.get("children", [])
+
+        try:
+            # need to check in children that had been previously registered
+            # TODO: this is where, once you've got the local children you run a
+            # predefined health check. online != healthy
+            for remote_child in remote_children:
+                # if the remote_child is not in the local_children, then it is offline
+                found = False
+                for local_child in local_children:
+                    if remote_child["slug"] == local_child.slug:
+                        is_available = True
+                        if remote_child["activeReservation"] is not None:
+                            if (
+                                remote_child["activeReservation"]["id"]
+                                and remote_child["isAvailable"]
+                            ):
+                                is_available = False
+
+                        self.primitive.hardware.check_in_http(
+                            is_available=is_available,
+                            is_healthy=True,
+                            is_online=True,
+                            fingerprint=remote_child["fingerprint"],
+                        )
+                        found = True
+                        break
+                if not found:
+                    logger.info(
+                        f"Remote child {remote_child['slug']}, not found in local children. Setting offline."
+                    )
+                    self.primitive.hardware.check_in_http(
+                        is_available=False,
+                        is_healthy=False,
+                        is_online=False,
+                        fingerprint=remote_child["fingerprint"],
+                    )
+
+        except Exception as exception:
+            logger.exception(f"Error checking in children: {exception}")
