@@ -44,6 +44,20 @@ def get_artifacts_cache(cache_id: str = None) -> Path:
     return artifacts_dir
 
 
+def get_logs_cache(cache_id: str = None) -> Path:
+    cache_dir = get_cache_dir()
+
+    logs_dir = cache_dir / "logs"
+
+    if cache_id:
+        logs_dir = logs_dir / cache_id
+
+    if not logs_dir.exists():
+        logs_dir.mkdir(parents=True, exist_ok=True)
+
+    return logs_dir
+
+
 def get_deps_cache() -> Path:
     cache_dir = get_cache_dir()
 
